open Lexing

type lexbuf = Lexing.lexbuf
type lex_tables = Lexing.lex_tables

let short tbl n =
  let w = 
    (Char.code (String.unsafe_get tbl (2*n))) + 
    ((Char.code (String.unsafe_get tbl (2*n + 1))) lsl 8)
  in
  if w < 0x8000 then w else w - 0x10000

let engine_classify_fun classify_fun tbl start_state lexbuf =
  lexbuf.lex_start_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
  lexbuf.lex_last_action <- (-1);
  
  let rec aux state =
    let base = short tbl.lex_base state in
    if base < 0 then (-base - 1)
    else
      let backtrk = short tbl.lex_backtrk state in
      if backtrk >= 0 then 
	(lexbuf.lex_last_pos <- lexbuf.lex_curr_pos;
	 lexbuf.lex_last_action <- backtrk);
      let c = 
  	if lexbuf.lex_curr_pos >= lexbuf.lex_buffer_len then
	  if lexbuf.lex_eof_reached then 0 
          else (
	    lexbuf.refill_buff lexbuf;
            if lexbuf.lex_eof_reached then 0 else classify_fun lexbuf
	  )
	else classify_fun lexbuf in
      let state = 
      	if short tbl.lex_check (base + c) = state then
	  short tbl.lex_trans (base + c)
	else
	  short tbl.lex_default state in
      if state < 0 then
	(lexbuf.lex_curr_pos <- lexbuf.lex_last_pos;
	 lexbuf.lex_last_action)
      else
	(if c = 0 then lexbuf.lex_eof_reached <- false;
	 aux state)

  in
  try aux start_state
  with End_of_file -> failwith "Unexpected end of stream during lexing"

let cur_char lexbuf =
  let b = Char.code (String.unsafe_get lexbuf.lex_buffer lexbuf.lex_curr_pos) 
  in
  lexbuf.lex_curr_pos <- lexbuf.lex_curr_pos + 1;
  b

let get_and_refill lexbuf =
  if (lexbuf.lex_curr_pos >= lexbuf.lex_buffer_len) then
    if lexbuf.lex_eof_reached then raise End_of_file
    else (
      lexbuf.refill_buff lexbuf;
      if (lexbuf.lex_curr_pos >= lexbuf.lex_buffer_len) then raise End_of_file
    );
  cur_char lexbuf


let engine_8bit table =
  assert (Array.length table >= 256);
  engine_classify_fun (fun lexbuf ->
	    let c = cur_char lexbuf in
	    Array.unsafe_get table c)


let engine_tiny_8bit table =
  assert (String.length table >= 256);
  engine_classify_fun (fun lexbuf ->
	    let c = cur_char lexbuf in
	    Char.code (String.unsafe_get table c))



let get_utf8_aux lexbuf =
  let c = get_and_refill lexbuf in
  if c >= 0x80 && c < 0xC0 then c
  else failwith "Invalid UTF-8 stream"
    
let get_utf8 lexbuf =
  let c1 = cur_char lexbuf in
  if c1 < 0x80 then c1 
  else if c1 >= 0xC0 && c1 < 0xE0 then
    let c2 = get_utf8_aux lexbuf in
    ((c1 - 0xC0) lsl 6) lor 
    ((c2 - 0x80))
  else if c1 >= 0xE0 && c1 < 0xF0 then
    let c2 = get_utf8_aux lexbuf in
    let c3 = get_utf8_aux lexbuf in
    ((c1 - 0xE0) lsl 12) lor 
    ((c2 - 0x80) lsl 6) lor
    ((c3 - 0x80))
  else if c1 >= 0xF0 && c1 < 0xF8 then
    let c2 = get_utf8_aux lexbuf in
    let c3 = get_utf8_aux lexbuf in
    let c4 = get_utf8_aux lexbuf in
    ((c1 - 0xF0) lsl 18) lor 
    ((c2 - 0x80) lsl 12) lor 
    ((c3 - 0x80) lsl 6) lor
    ((c4 - 0x80))
  else failwith "Invalid UTF-8 stream"


let engine_tiny_utf8 table def =
  engine_classify_fun (fun lexbuf ->
	    let c = get_utf8 lexbuf in
	    if c >= String.length table then def c
	    else Char.code (String.unsafe_get table c))

